//
// Copyright (c) 1997,2000 by the University of Southern California
// All rights reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation in source and binary forms for non-commercial purposes
// and without fee is hereby granted, provided that the above copyright
// notice appear in all copies and that both the copyright notice and
// this permission notice appear in supporting documentation. and that
// any documentation, advertising materials, and other materials related
// to such distribution and use acknowledge that the software was
// developed by the University of Southern California, Information
// Sciences Institute.  The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
//
// THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
// the suitability of this software for any purpose.  THIS SOFTWARE IS
// PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// Other copyrights might apply to parts of this software and are so
// noted when applicable.
//
// $Header: /nfs/jade/vint/CVSROOT/nam-1/enetmodel.cc,v 1.17 2000/03/24 03:49:04 haoboy Exp $
//
// Network model with Editor layout

#include <stdlib.h>
#include <math.h>
#include <float.h>

#include "random.h"
#include "view.h"
#include "netview.h"
#include "animation.h"
#include "queue.h"
#include "edge.h"
#include "node.h"
#include "agent.h"
#include "sincos.h"
#include "state.h"
#include "packet.h"
#include "enetmodel.h"

static int agent_number = 0;

class EditorNetworkModelClass : public TclClass {
public:
	EditorNetworkModelClass() : TclClass("NetworkModel/Editor") {}
	TclObject* create(int argc, const char*const* argv) {
		if (argc < 5) 
			return 0;
		return (new EditorNetModel(argv[4]));
	}
} editornetworkmodel_class;

EditorNetModel::EditorNetModel(const char *editor) : NetModel(editor)
{
	nID = 0;

	bind("Wpxmin_", &pxmin_);
        bind("Wpymin_", &pymin_);
	bind("Wpxmax_", &pxmax_);
        bind("Wpymax_", &pymax_);
}

EditorNetModel::~EditorNetModel()
{
}

void EditorNetModel::BoundingBox(BBox& bb)
{                       
	// by default, 800X1000 internal drawing area
        bb.xmin = pxmin_;
	bb.ymin = pymin_;
        bb.xmax = pxmax_;
        bb.ymax = pymax_;

        for (Animation* a = drawables_; a != 0; a = a->next())
                a->merge(bb);

	pxmin_ = bb.xmin;
	pymin_ = bb.ymin;
	pxmax_ = bb.xmax;
	pymax_ = bb.ymax;
}

int EditorNetModel::addNode(float cx, float cy)
{
       /*
        * <net> node <name> <shape> <color> <addr> [<size>]
        * Create a node using the specified name
        * and the default size and insert it into this
        * NetModel's list of drawables.
        */
	 Node *n;
	 static char name[TRACE_LINE_MAXLEN];

	 if (nID == 0 ) {

	     for (n = nodes_; n != 0; n = n->next_) {

	   	if((n->num())>nID) nID = n->num();	 
	     }

	     if (nID > 0) nID ++;

	 }
	 
	 sprintf(name, "%d", nID);

         double size = 70;

         n = new CircleNode(name,size);     
	 n->init_color("black");
	 nID ++;

         //n->setaddr(addr);
         //addAddress(n->num(), addr);
         n->place(cx,cy);
         n->next_ = nodes_;
         nodes_ = n;
         n->insert(&drawables_);
         return (TCL_OK);
}

int EditorNetModel::addLink(Node *src, Node *dst)
{
	double bw = 1000000.0;
	double delay = 0.002;
	double length = 0;
	double angle = 1.0;

	Edge *e = new Edge(src, dst, 25,
			   bw, delay, length, angle);
        e->init_color("black");
        enterEdge(e);
        e->insert(&drawables_);
        src->add_link(e);
//        tcl.resultf("%g", delay);

	// duplex link by default according to nam
	e = new Edge(dst, src, 25,  
		     bw, delay, length, angle);

	e->init_color("black");
	enterEdge(e);
	e->insert(&drawables_);
        dst->add_link(e);
	
        return (TCL_OK);
}

int EditorNetModel::addAgent(Node *src, char *agent, float cx, float cy)
{
        Node *n = lookupNode(src->num());
	Agent *a;

	if (n==0) {
	  return 0;
	} else {
	  a = new BoxAgent(agent, n->size());
	  placeAgent(a, src);
	  a->insert(&animations_);
	  a->node_ = n;
	  a->number_ = agent_number++;
	  n->add_agent(a);
	}
	return (TCL_OK);
}

int EditorNetModel::addAgentLink(Agent *src_agent, Agent *dst_agent)
{
         float src_x_, src_y_, dst_x_, dst_y_;
         src_x_ = src_agent->x(), src_y_ = src_agent->y();
         dst_x_ = dst_agent->x(), dst_y_ = dst_agent->y();

         if ((strcmp(src_agent->name(), "TCPSink")==0)|| \
	     (strcmp(src_agent->name(), "TCPSink/DelAck")==0)|| \
	     (strcmp(src_agent->name(), "TCPSink/Sack1")==0)) {
	            src_agent->AgentRole_ = DESTINATION;
	            dst_agent->AgentRole_ = SOURCE;
	 } else {
	            src_agent->AgentRole_ = SOURCE;
	            dst_agent->AgentRole_ = DESTINATION;
	 }
	 
	 dst_agent->AgentPartner_ = src_agent;
	 src_agent->AgentPartner_ = dst_agent;

	 return (TCL_OK);
}

void EditorNetModel::setNodeProperty(int id, char* pv, int pn)
{
	Node *n = lookupNode(id);
        Agent *a;
	if (n == NULL) {
            fprintf(stderr, "Nonexisting node %d.\n", id);
	    return;
        }

	switch (pn) {
	    case 0:
	    {
		double s = atof(pv);
		n->size(s);
		break;
	    }
	    case 1:	//color
	    {
		n->init_color(pv);
		break;
	    }
	    case 2:	//label
	    {
       	         if (strcmp(pv, "(null)") != 0) 
	                 n->dlabel(pv);
		 break;
	    }
        }
}

void EditorNetModel::setAgentProperty(int id, char* pv, int pn)
{
	Agent *a = lookupAgent(id);
        if (a == NULL) {
          fprintf(stderr, "Nonexisting agent %s.\n", id);
          return;
        }

	switch (pn) {
	     case 0:    // initial window size
	       {
		    int size = atoi(pv);
		    a->windowInit(size);
		    break;
	       }
	     case 1:    // window size
	       {
		   int size = atoi(pv);
		   a->window(size);
		   break;
	       }
	     case 2:     // maximum cwnd_
	       {
		   int size = atoi(pv);
		   a->maxcwnd(size);
		   break;
	       }
	     case 3:     // tracevar
	       {
		   a->tracevar(pv);
		   break;
		}
	     case 4:     // start-at
	       {
		 float time = atof(pv);
		 a->startAt(time);
		 break;
	       }
	     case 5:     // FTP producemore
	       {
		 int size = atoi(pv);
		 a->produce(size);
		 break;
	       }
             case 6:     // stop-at
	       {
		 float time = atof(pv);
		 a->stopAt(time);
		 break;
	       }
	     case 7:    // Flow ID + Color
	       {
	          if ((strcmp(pv, "(null)") == 0) || (strcmp(pv, "") == 0)) {
		            pv = "black";
			    a->flowcolor(pv);
		  } else {
		            a->flowcolor(pv);
		  } 
		 break;
	       }
	     case 8:    // CBR packetSize
	       {
		 int size = atoi(pv);
		 a->packetSize(size);
		 break;
	       }
	     case 9:    // CBR interval
	       {
		 float size = atof(pv);
		 a->interval(size);
		 break;
	       }

	}
}

void EditorNetModel::setLinkProperty(int sid,int did, char* pv, int pn)
{
	EdgeHashNode *h = lookupEdge(sid, did);
	EdgeHashNode *g = lookupEdge(did, sid);

        if (h == NULL || g == NULL ) {
            fprintf(stderr, "Nonexisting link %d.%d\n", sid, did);
            return;
        }

        switch (pn) {
            case 0: //bandwidth
            {
                double s = atof(pv);
		(h->edge) -> setBW(s);
		(g->edge) -> setBW(s);
                break; 
            }
            case 1:     //color
            {
                //int color = atoi(pv);
                (h->edge)->init_color(pv);
                (g->edge)->init_color(pv);
                break;
            }
            case 2:     //delay
            {
                double s = atof(pv);
		(h->edge) -> setDelay(s);
		(g->edge) -> setDelay(s);
                break;
            }
	    case 3:     //label
	      {
		if (strcmp(pv, "(null)") != 0) {
		        (h->edge) -> dlabel(pv);
			(g->edge) -> dlabel(pv);
		}
		break;
	      }
	     default: break;
        }
}

int EditorNetModel::command(int argc, const char *const *argv)
{               
        if (argc == 2) {
		if (strcmp(argv[1], "layout") == 0) {
			layout();
			return (TCL_OK);
	        }
	}
        if (argc == 3) {
                if (strcmp(argv[1], "saveasenam") == 0) {
                        /*
                         * <net> saveasenam filehandler
                         */
                        saveAsEnam(argv[2]);
                        return (TCL_OK);
                }
                if (strcmp(argv[1], "saveasns") == 0) {
                        /*
                         * <net> saveasns filehandler
                         */
                        saveAsNs(argv[2]);
                        return (TCL_OK);
                }
        }       
        return (NetModel::command(argc, argv));
}                  

void EditorNetModel::layout()
{
        Node *n;
        for (n = nodes_; n != 0; n = n->next_)
                for (Edge* e = n->links(); e != 0; e = e->next_)
                        placeEdge(e, n);
}

// XXX losing some editor's info

int EditorNetModel::saveAsNs(const char *filename)
{

  FILE *file;
  Node *n;
  Edge *e;      
  Agent *a;
  int ret;
  file=fopen(filename, "a");
  if (file==0) 
  {     
    fprintf(stderr, "nam: Couldn't open file: %s\n", filename);
    return -1;
  }

  fprintf(file, "\n#create ns\n");
  fprintf(file, "set ns [new Simulator]\n");
  fprintf(file, "$ns namtrace-all [open out.nam w]\n");

  // save node info
   fprintf(file, "\n#node creation\n");

   if (!nodes_) return 0;
   int i = nodes_->num();
   int j = nodes_->next_->num();

   // for consistent numbering between nam editor and nam
   if (i > j) {
        for (int k = 0; k <= i; k++) {
            for (n = nodes_; n != 0; n = n->next_) {
                 if (k==n->num())   { 
	                ret = n->saveAsNs(file);
                        if (ret!=0) {
	                fclose(file);
	                return -1;
	                }
                  fprintf(file, "\n");
                  }
	    }
        }
   } else {
       for (n = nodes_; n != 0; n = n->next_) {
	    ret = n->saveAsNs(file);
	    if (ret!=0) {
	        fclose(file);
	        return -1;
	    }
	 fprintf(file, "\n");
       }
   }

  // save link info
  fprintf(file, "#link creation\n");
  for (n = nodes_; n != 0; n = n->next_)
    for(e= n->links(); e !=0; e = e->next_)
    {
      ret = e->saveAsNs(file);
      if (ret!=0) {
        fclose(file);
        return -1;
      }
    }

  // save agent info
  fprintf(file, "#agent creation\n");
  for (n = nodes_; n != 0; n = n->next_) {
    for(a= n->agents(); a !=0; a = a->next_) {
	ret = a->saveAsNs(file);
    }
  }
  for (n = nodes_; n != 0; n = n->next_) {
    for(a= n->agents(); a !=0; a = a->next_) {
	if (a->AgentRole_ == SOURCE) {
	  if (strcmp(a->name(), "CBR") == 0) { // CBR Traffic
	          int src;
		  src = a->number_;
		  fprintf(file, "$ns connect $agent%d $agent%d \n\n", src, a->AgentPartner_->number_);
		  fprintf(file, "$ns at %f \"$agent%d start\" \n", a->startAt(), src);
		  fprintf(file, "$ns at %f \"$agent%d stop\" \n\n", a->stopAt(), src);

	   } else {                   // TCP traffic
	          int src;
		  src = a->number_;
		  fprintf(file, "$ns connect $agent%d $agent%d \n\n", src, a->AgentPartner_->number_);
		  fprintf(file, "\nset ftp%d [new Application/FTP] \n", src);
		  fprintf(file, "$ftp%d attach-agent $agent%d \n\n", src, src);
		  if (a->produce()>0) {          // ftp producemore option 
		       fprintf(file, "$ns at %f \"$ftp%d producemore %d\" \n", 
			              a->startAt(), src, a->produce());
		  } else {
		       fprintf(file, "$ns at %f \"$ftp%d start\" \n", a->startAt(), src);
		       fprintf(file, "$ns at %f \"$ftp%d stop\" \n\n", a->stopAt(), src);
		  }
	   }
	}
    }
  }

  fprintf(file, "#Run\n");
  fprintf(file, "exec nam out.nam &\n");
  fprintf(file, "$ns run\n");
  return(fclose(file));

}

int EditorNetModel::saveAsEnam(const char *filename)
{

  FILE *file;   
  Node *n;
  Edge *e;      
  Agent *a;
  int ret;       
  file=fopen(filename, "w");
  if (file==0)   
  {
    fprintf(stderr, "nam: Couldn't open file: %s\n", filename);
    return -1;  
  }             

  // save warning info
  fprintf(file, "## All lines starting with ## are created by nam editor.\n");
  fprintf(file, "## Please do not try to edit them manually.\n");

  // save version info
  fprintf(file, "##V -t * -v 1.0a7 -a 0\n");

  // save page size
  fprintf(file, "##W -t * -x %f -y %f -X %f -Y %f\n",pxmin_, 
	pymin_, pxmax_, pymax_);

  // save node information
  for (n = nodes_; n != 0; n = n->next_)
  {   
    ret = n->saveAsEnam(file); 
    if (ret!=0) {
      fclose(file);
      return -1;
    }           
  }

  //save link information
  for (n = nodes_; n != 0; n = n->next_)
    for(e= n->links(); e !=0; e = e->next_)
    {           
      ret = e->saveAsEnam(file);
      if (ret!=0) {
        fclose(file);   
        return -1;
      }
    } 

  // save agent information
  for (n = nodes_; n != 0; n = n->next_) 
    for(a= n->agents(); a !=0; a = a->next_) 
      {
         ret = a->saveAsEnam(file);
	 if (ret!=0) {
	   fclose(file);
	   return -1;
	 }
      }

  fprintf(file, "## Please do not try to edit this file above this line.\n\n");
  return(fclose(file)); 

}

void EditorNetModel::removeNode(Node *n)
{
	Node *p,*q;
	p = nodes_;     

	for (q = nodes_; q != 0; q = q->next_) {
		if ( n->num() == q->num() ) {
			// when deleting the last node
			if ( q->num() == nodes_->num() ) {
				nodes_=nodes_->next_;
				break;
			} else {
				p->next_ = q->next_; 
				break;
			}
		}
		p = q;
	}
	q->detach();  

	// delete edge on the nodes_
	for(Edge *e= q->links(); e !=0; e = e->next_) 
		removeLink(e);

	// delete agents on the nodes_
	for(Agent *a= q->agents(); a !=0; a = a->next_) 
		removeAgent(a);

	delete q;
}

void EditorNetModel::removeLink(Edge *e)
{
	int src = e->src();
	int dst = e->dst();

	EdgeHashNode *h = lookupEdge(src, dst);
	EdgeHashNode *g = lookupEdge(dst, src);
	if (h == 0 || g == 0) {
		// h,g = 0 or h,g !=0
		return;
	} 

	Edge* e1 = h->edge;
	Edge* e2 = g->edge;
 
	removeEdge(e1);
	removeEdge(e2);
 
	e1->detach();
	e2->detach();
 
	Node* srcnode = e1->start();
	Node* dstnode = e2->start();

	// it is a duplex by default
	srcnode->delete_link(e1);
	dstnode->delete_link(e2); 

	delete e1;
	delete e2;
}

void EditorNetModel::removeAgent(Agent *a)
{
	Node *n;
	n = a->node_;
	
	if (a->AgentPartner_!=NULL) {
		a->AgentPartner_->AgentPartner_ = NULL;
		a->AgentPartner_->AgentRole_ = 0;
	}
	n->delete_agent(a);
	a->detach();
	delete a;
}


